#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ncurses.h>
#include <sys/time.h>
#include <getopt.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>
#include <fcntl.h>

#include <lq.h>
#include <config.h>


#define MODE_NORMAL	1
#define MODE_SELECT	2

struct cell {
	lq_stats	stats;
	unsigned int	load;
	unsigned int	version;

	char		*name;

	struct cell	*next;
};

static struct cell *R = NULL, *selected_cell = NULL;
static unsigned int names_mtime = 0;
static unsigned int version = 0;
static unsigned int classes;
static unsigned int page = 0;
static unsigned int max_pages = 1;
static unsigned int selected_line = 1;
static unsigned int mode = MODE_NORMAL;
static unsigned long filter_parent = 0;
static unsigned int max_line_len;

struct option options[] =
{
	{"interface", required_argument, NULL, 'i'},
	{"refresh", required_argument, NULL, 'r'},
	{"server", required_argument, NULL, 's'},
	{"port", required_argument, NULL, 'p'},
	{"names", required_argument, NULL, 'n'},
	{"help", no_argument, NULL, 'h'},
	{NULL, 0, NULL, 0}
};


// Functions
static void lq_set_class_name(unsigned int id, char *name)
{
	struct cell *p;

	p = R;
	while (p) {
		if (p->stats.handle == id) {
			if (p->name != NULL)
				free(p->name);
			p->name = strdup(name);
			break;
		}

		p = p->next;
	}
}

static void lq_dump2file(char *dev)
{
	struct cell *p;
	lq_stats *s;
	int out;
	char file[128];
	char buf[1024];

	snprintf(file, sizeof(file), "lq_%s_%ld.dump",
		dev, time(NULL));

	out = open(file, O_WRONLY | O_CREAT | O_EXCL | O_TRUNC, 0600);
	if (out == -1)
		return;

	p = R;
	while (p) {
		s = &p->stats;
		snprintf(buf, sizeof(buf), "%4x:%4x %32s parent=%4x:%4x bps=%u pps=%u drops=%u overlimits=%u load=%u qlen=%u rate=%ubps ceil=%ubps\n",
			s->handle >> 16, s->handle & 0xffff, p->name ? p->name : "?",
			s->parent >> 16, s->parent & 0xffff,
			s->bps, s->pps, s->drops, s->overlimits, p->load, s->qlen, s->rate, s->ceil);
		write(out, buf, strlen(buf));

		p = p->next;
	}

	close(out);
}

static int lq_load_names(char *file, unsigned int *line_with_error)
{
	FILE *f;
	struct stat S;
	int ret;
	char *r, buf[256], *c;
	unsigned long id;
	unsigned int line, len;

	ret = stat(file, &S);
	if (ret == -1) {
		*line_with_error = 0;
		return 0;
	}

	// need to reopen?
	if (names_mtime == S.st_mtime)
		return 0;

	f = fopen(file, "r");
	if (!f)
		return -1;

	*line_with_error = 0;
	line = 0;
	while (1) {
		r = fgets(buf, sizeof(buf), f);
		if (!r)
			break;
		line++;
		if (buf[0] == '#')
			continue;

		len = strlen(buf);

		while ((buf[len - 1] == '\n') || (buf[len - 1] == '\r')
			|| (buf[len - 1] == '\t') || (buf[len - 1] == ' ')) {
			buf[len - 1] = '\0';
			len--;
		}

		if (len == 0)
			continue;

		id = strtoul(r, &r, 16);
		id <<= 16;

		if (*r != ':') {
			if (*line_with_error == 0)
				*line_with_error = line;
			continue;
		}

		r++;
		id |= strtoul(r, &r, 16);

		while ((*r == ' ') || (*r == '\t') || (*r == '^'))
			r++;

		c = strchr(r, '^');
		if (c)
			*c = '\0';

		len = strlen(r);
		if (len == 0) {
			if (*line_with_error == 0)
				*line_with_error = line;
			continue;
		}

		lq_set_class_name(id, r);
	}

	fclose(f);

	return 0;
}

static void usage(void)
{
	printf("lq - Author: Catalin(ux aka Dino) BOIE\n");
	printf("Usage:\n");
	printf("	--interface (-i) <interface> [eth0]\n");
	printf("	--refresh (-r) <seconds> [1]\n");
	printf("	--names (-n) <file_with_id_tab_names> [/etc/lq/<dev>.txt]\n");
	printf("	--server (-s) <QOS server address> []\n");
	printf("	--port (-p) <QOS server port> [65000]\n");
	printf("	--help (-h) - this help\n");

	printf("\n");
}

static char *get_sort_by(unsigned int id)
{
	static char ret[64], *r;
	unsigned int x, inverse;

	x = id & 0xff;
	inverse = id >> 8;

	switch (id) {
		case 1: r = "bps"; break;
		case 2: r = "pps"; break;
		case 3: r = "qlen"; break;
		case 4: r = "drops"; break;
		case 5: r = "load"; break;
		case 6: r = "overlimit"; break;
	}

	snprintf(ret, sizeof(ret), "%s%s",
		r, inverse == 1 ? " - inv" : "");

	return ret;
}

/*
	by: 1 - bps, 2 - pps, 3 - qlen, 4 - drops, 5 - load, 6 - overlimit
*/
static void sort_cells(unsigned int by_inverse)
{
	struct cell *p, *n, *prev;
	int stop = 0;
	u_int64_t v, v1 = 0, v2 = 0;
	unsigned int by, inverse, switches = 0;

	by = by_inverse & 0xff;
	inverse = by_inverse >> 8;

	while (!stop) {
		stop = 1;
		p = R;
		prev = NULL;
		while (p) {
			n = p->next;
			if (!n)
				break;

			switch (by) {
				case 1: v1 = p->stats.bps; v2 = n->stats.bps; break;
				case 2: v1 = p->stats.pps; v2 = n->stats.pps; break;
				case 3: v1 = p->stats.qlen; v2 = n->stats.qlen; break;
				case 4: v1 = p->stats.drops; v2 = n->stats.drops; break;
				case 5: v1 = p->load; v2 = n->load; break;
				case 6: v1 = p->stats.overlimits; v2 = n->stats.overlimits; break;
			}

			if (inverse == 1) {
				v = v1;
				v1 = v2;
				v2 = v;
			}

			if (v1 < v2) {
				if (prev)
					prev->next = n;
				else
					R = n;
				p->next = n->next;
				n->next = p;
				stop = 0;
				switches++;
			}

			prev = p;
			p = p->next;
		}
	}
}

static void show_cells(unsigned int page, unsigned int classes)
{
	char buf[256], bps[64], rate[64], ceil[64], id[20];
	struct cell *p;
	lq_stats *s;
	int limit, max, line, skip;
	unsigned int no_of_lines;

	max = max_line_len;
	if (max > sizeof(buf))
		max = sizeof(buf);

	// limit is the number of lines we can show
	limit = LINES - 3;

	// skip is the number of entries we must skip
	skip = limit * page;

	no_of_lines = classes - skip;

	if (selected_line > no_of_lines)
		selected_line = no_of_lines;

	p = R;
	line = 2;
	while (p) {
		if (skip > 0) {
			skip--;
			p = p->next;
			continue;
		}

		s = &p->stats;

		lq_rate2string(s->bps * 8, bps, sizeof(bps));
		lq_rate2string(s->rate * 8, rate, sizeof(rate));
		lq_rate2string(s->ceil * 8, ceil, sizeof(ceil));

		if (p->name == NULL)
			snprintf(id, 13, "%4x:%4x", s->handle >> 16, s->handle & 0xffff);
		else
			snprintf(id, 13, "%9s", p->name);

		snprintf(buf, max, "%12s %10s/s %7u %5u%% %4u %8u %8u %10s/s %10s/s",
			id,
			bps, s->pps,
			p->load,
			s->qlen,
			s->drops, s->overlimits,
			rate, ceil);

		if ((mode == MODE_SELECT) && (line == selected_line + 1)) {
			selected_cell = p;
			attron(COLOR_PAIR(4));
		} else {
			attron(COLOR_PAIR(2));
		}
		mvaddstr(line, 0, buf);

		limit--;
		if (limit == 0)
			break;

		line++;
		p = p->next;
	}
}

static unsigned int clean_old_cells(void)
{
	struct cell *prev, *p, *next;
	unsigned int ret = 0;

	p = R;
	prev = NULL;
	while (p != NULL) {
		ret++;
		next = p->next;

		if (p->version != version) {
			// take care if current class is deleted
			if (selected_cell && (p->stats.handle == selected_cell->stats.handle))
				selected_cell = NULL;

			if (prev)
				prev->next = next;
			else
				R = next;
			free(p);
		} else {
			prev = p;
		}
		p = next;
	}

	return ret;
}

static void cb(lq_stats *s)
{
	struct cell *p, *q = NULL, *prev;
	int found;

	if (filter_parent && (filter_parent != s->parent))
		return;

	classes++;

	found = 0;
	p = R;
	prev = NULL;
	while (p) {
		if (p->stats.handle == s->handle) {
			q = p;
			found = 1;
			break;
		}

		prev = p;
		p = p->next;
	}

	if (found == 0) {
		q = malloc(sizeof(struct cell));
		if (!q)
			return;

		memset(q, 0, sizeof(struct cell));

		if (prev)
			prev->next = q;
		else
			R = q;
	}

	// common stuff
	q->version = version;
	memcpy(&q->stats, s, sizeof(lq_stats));

	// do some computing
	q->load = 0;
	if (q->stats.rate > 0)
		q->load = 100 * q->stats.bps / q->stats.rate;
}

static void lq_show_help(void)
{
	char buf[256];
	int ch, max;

	max = max_line_len;
	if (max > sizeof(buf))
		max = sizeof(buf);

	// wait for a key when calling getch
	nodelay(stdscr, FALSE);

	erase();

	attron(COLOR_PAIR(2));

	snprintf(buf, max, "Help page - (C)opyright Catalin(ux aka Dino) BOIE 2005\n\n");
	mvaddstr(0, 0, buf);

	// some info
	snprintf(buf, max,
		"Official website: http://kernel.umbrella.ro/\n"
		"\n");
	addstr(buf);

	// sorting
	snprintf(buf, max, "Sorting:\n"
		"	0 - Reverse sorting		4 - by drops\n"
		"	1 - by bps			5 - by load\n"
		"	2 - by pps			6 - by overlimit\n"
		"	3 - by qlen\n"
		"\n");
	addstr(buf);

	// selection
	snprintf(buf, max, "Selection / filtering\n"
		"	s - Enter/exit in/from selection mode\n"
		"	Up/Down - enter in selection mode\n"
		"	c - Filter classes that have the current class as parent\n"
		"	p - Filter classes that have the same parent as current class\n"
		"	u - Unfilter\n"
		"\n");
	addstr(buf);

	// navigation
	snprintf(buf, max, "Navigation:\n"
		"	Up/Down - one line up/down\n"
		"	PgUp/Down - go to previous/next page\n"
		"	Home/End - go to first/last page\n"
		"\n");
	addstr(buf);

	// press a key for next page
	snprintf(buf, max, "Press any key for next page.\n");
	mvaddstr(LINES - 1, 0, buf);

	refresh();
	ch = getch();

	// page 2

	erase();

	// misc
	snprintf(buf, max, "Misc:\n"
		"	h - help\n"
		"	q - quit program\n"
		"	d - dump list to file (lq_<interface>_<timestamp>.dump)\n"
		"\n");
	addstr(buf);

	// names file
	snprintf(buf, max, "Names file (--names switch):\n"
		"	The format is: <classid> <name>\n"
		"	Example:\n"
		"		1:1 ROOT\n"
		"		1:2 Class1\n"
		"	Default file is /etc/lq/<interface>.txt\n"
		"\n\n");
	addstr(buf);

	// end
	snprintf(buf, max, "Press any key to exit help.\n");
	mvaddstr(LINES - 1, 0, buf);

	refresh();
	ch = getch();

	nodelay(stdscr, TRUE);
}

int main(int argc, char *argv[])
{
	int idev;
	int ret;
	int ch, key_pressed;
	unsigned int sort_by = 1, inverse = 0, re_sort = 0, force_sort;
	char s_sort[256];
	struct timeval tdump_s, tdump_e;
	long tdump = 0;
	char stdump[40], status[256];
	char *dev = "eth0";
	int c;
	unsigned long refresh_rate = 1;
	int port = 65000;
	char *server;
	char *names = NULL;
	char names_file[1024];
	char load_status[128];
	unsigned int line = 0;
	unsigned int refresh_count = 0, refresh_steps;
	struct timespec ts;
	unsigned int force_exit = 0;
	char error[256];
	int errcode;

	while (1) {
		int option_index = 0;

		c = getopt_long(argc, argv, "i:r:s:p:n:h", options, &option_index);
		if (c == -1)
			break;

		switch (c) {
			case 'i': dev = optarg; break;
			case 'r': refresh_rate = strtoul(optarg, NULL, 10); break;
			case 's': server = optarg; break;
			case 'p': port = strtoul(optarg, NULL, 10); break;
			case 'n': names = optarg; break;
			case 'h': usage(); return 0;

			case '?':
				printf("Invalid option!\n");
				usage();
				return 1;
		}
	}

	if (names == NULL) {
		snprintf(names_file, sizeof(names_file), "%s/lq/%s.txt",
			LQ_ETC_DIR, dev);
		names = names_file;
	}

	lq_debug = 0;

	ret = lq_init();
	if (ret == -1) {
		printf("Cannot init lq [%s]\n",
			lq_strerror());
		return 1;
	}

	idev = lq_getif(dev);
	if (idev == -1) {
		printf("Error geting info about interface %s [%s]!\n",
			dev, lq_strerror());
		return 1;
	}


	initscr();
	start_color();
	cbreak();
	noecho();

	nonl();
	intrflush(stdscr, FALSE);
	keypad(stdscr, TRUE);
	nodelay(stdscr, TRUE);


	init_pair(1, COLOR_WHITE, COLOR_BLACK);
	init_pair(2, COLOR_WHITE, COLOR_BLACK);
	init_pair(3, COLOR_WHITE, COLOR_RED);
	init_pair(4, COLOR_RED, COLOR_WHITE);


	refresh_steps = refresh_rate * 20;

	force_sort = 1;

	strcpy(error, "");
	errcode = 0;

	while (1) {
		max_line_len = COLS + 1;
		if (max_line_len > sizeof(status))
			max_line_len = sizeof(status);

		key_pressed = 0;
		if ((ch = getch()) != ERR) {
			key_pressed = 1;
			re_sort = 0;
			switch (ch) {
				case '0':
				case '1':
				case '2':
				case '3':
				case '4':
				case '5':
				case '6':
				case '7':
				case '8':
				case '9': re_sort = 1; break;
				case 'd': lq_dump2file(dev); break;
				case 'h': lq_show_help(); break;
				case 'q': force_exit = 1;
				case 's':
					if (mode == MODE_SELECT)
						mode = MODE_NORMAL;
					else
						mode = MODE_SELECT;
					break;
				case 'c': /* current as parent */
					if ((mode == MODE_SELECT) && selected_cell)
						filter_parent = selected_cell->stats.handle;
					break;
				case 'p':
					if ((mode == MODE_SELECT) && selected_cell)
						filter_parent = selected_cell->stats.parent;
					break;
				case 'u': filter_parent = 0; break;
				case KEY_NPAGE:
					if (page + 1 < max_pages) {
						page++;
						selected_line = 1;
					}
					break;
				case KEY_PPAGE:
					if (page > 0) {
						page--;
						selected_line = 1;
					}
					break;
				case KEY_HOME:
					page = 0;
					selected_line = 1;
					break;
				case KEY_END:
					page = max_pages - 1;
					selected_line = 1;
					break;

				case KEY_UP:
					if (mode == MODE_NORMAL) {
						mode = MODE_SELECT;
					} else {
						if (selected_line > 1) {
							selected_line--;
						} else if (page > 0) {
							page--;
							selected_line = LINES - 3;
						}
					}
					break;
				case KEY_DOWN:
					if (mode == MODE_NORMAL) {
						mode = MODE_SELECT;
					} else {
						if (selected_line < LINES - 3) {
							selected_line++;
						} else if (page + 1 < max_pages) {
							page++;
							selected_line = 1;
						}
					}
					break;

				default:
					key_pressed = 0;
			}

			if (force_exit == 1)
				break;
		}

		// sorting
		if ((force_sort == 1) || (re_sort == 1)) {
			force_sort = 0;
			re_sort = 0;

			switch (ch) {
				case '0': inverse = 1 - inverse; break;
				case '1': sort_by = 1; break;
				case '2': sort_by = 2; break;
				case '3': sort_by = 3; break;
				case '4': sort_by = 4; break;
				case '5': sort_by = 5; break;
			}

			snprintf(s_sort, sizeof(s_sort), "[Sort by: %s]",
				get_sort_by(sort_by | (inverse << 8)));
		}

		// test if is the momment to do the refresh
		if ((refresh_count == 0) || (key_pressed == 1)) {
			version = 1 - version;
			classes = 0;
			gettimeofday(&tdump_s, NULL);
			ret = lq_getstats(LQ_class, idev, cb);
			if (ret == -1) {
				snprintf(error, sizeof(error), "Error getting stats [%s]!\n",
					lq_strerror());
				errcode = 1;
				break;
			}
			gettimeofday(&tdump_e, NULL);
			tdump = (tdump_e.tv_sec - tdump_s.tv_sec) * 1000000 +
				tdump_e.tv_usec - tdump_s.tv_usec;
			snprintf(stdump, sizeof(stdump), "[Tdump:%ldms] [%d classes]",
				tdump / 1000, classes);

			classes = clean_old_cells();

			max_pages = classes / (LINES - 3) + 1;

			ret = lq_load_names(names, &line);
			if (ret == 0) {
				if (line != 0) {
					snprintf(load_status, sizeof(load_status),
						" Error loading names in line %d!",
						line);
				} else {
					snprintf(load_status, sizeof(load_status),
						"%s", "");
				}
			} else {
				snprintf(load_status, sizeof(load_status),
					" Error loading %s",
					names);
			}


			if (mode == MODE_NORMAL)
				sort_cells(sort_by | (inverse << 8));

			// draw
			erase();

			// up status
			attron(COLOR_PAIR(2));
			snprintf(status, max_line_len, "%s   (C)opyright Catalin(ux aka Dino) BOIE 2005   [Press h for help]\n",
				dev);
			mvaddstr(0, 0, status);

			// info
			attron(COLOR_PAIR(3));
			snprintf(status, max_line_len, "          ID        Speed     pps   load qlen    drops     over         rate         ceil\n");
			mvaddstr(1, 0, status);

			show_cells(page, classes);

			// down status
			attron(COLOR_PAIR(3));
			snprintf(status, max_line_len, "%s %s%s [Page %u/%u]",
				s_sort, stdump, load_status, page + 1, max_pages);
			mvaddstr(LINES - 1, 0, status);

			refresh();

			// reinit counter
			if (refresh_count == 0)
				refresh_count = refresh_steps;
		} else {
			refresh_count--;
		}

		ts.tv_sec = 0;
		ts.tv_nsec = 50000000;
		nanosleep(&ts, NULL);
	}

	endwin();

	if (errcode > 0)
		fprintf(stderr, "%s", error);

	return errcode;
}
